// mozileUndoRedo.js

/* ***** BEGIN LICENSE BLOCK *****
 * Licensed under Version: MPL 1.1/GPL 2.0/LGPL 2.1
 * Full Terms at http://mozile.mozdev.org/license.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Playsophy code.
 *
 * The Initial Developer of the Original Code is Playsophy
 * Portions created by the Initial Developer are Copyright (C) 2002-2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *		Karl Guertin <grayrest@grayrest.com>
 *		James A. Overton <james@overton.ca>
 *		Christian Stocker (Bitflux)
 *		Paul Everitt
 *		David Palm
 *
 * ***** END LICENSE BLOCK ***** */

/*******

Mozile Undo Redo Javascript
Mozile version 0.61

Created by James A. Overton
Created on 2004-03-24

Last revised by James A. Overton
Last revised on 2004-04-17

HISTORY:
	Some code in this file was first found in mozilehtmltb.js, while the 
majority was cut out of mozileTB.js and separated into different files.


DESCRIPTION:
	This code handles the mozileUndo and mozileRedo functions, as well as the
mozileDo function.

*********/




/********** Undo Functionality *************/

// This is the lazy way, I copy the whole BODY...
// This code still needs to be tweaked.


// Mozile Do
// Records the state of the current document into the undoStack for that document,
// and increments the undoCounter (each document gets its own stack and counter).
// All steps higher up in the stack are eliminated when a lower one is added.
// If the undoMax number of steps is hit, then the stack is spliced.
// The undoStack includes the name of the command and the whole BODY node.

// Yes, it is a brute force solution to copy the whole BODY,
// but it is the best solution for multiple contentEditable areas,
// (doing the whole HTML messes up the scrollbar)
// and the performance seems fine.
// Of course this won't work for XML in general...

const UNDO_NAME = 0;
const UNDO_NODE = 1;

function mozileDo(name) {
	var f = ["mozileUndoRedo.js","mozileDo"];
	mozileDebug(f,3,"Do: "+name);
	
	/*MAX: Is this buggy??? Indexing array with strings not allowed???*/
	try { 
		mozileTarget.document.MOZILE_CURRENT_CONFIG["saved"] = 'false';
	} catch(e) {	
		mozileTarget.document.MOZILE_CURRENT_CONFIG = new Array();
		mozileTarget.document.MOZILE_CURRENT_CONFIG["saved"] = 'false';
	}

	try{
		if( (("undoStack" in mozileTarget) != true) || (!mozileTarget.undoStack) ) {
			mozileTarget.undoStack=new Array();
			mozileTarget.undoCounter=0;
		}
		var undoStep = new Array;
		var getNode;
		mozileDebug(f,3,"Context: "+ mozileTarget.document.documentElement.tagName);
		if(mozileTarget.document.documentElement.tagName=="HTML") {
			mozileDebug(f,3,"HTML context");
			getNode = mozileTarget.document.getElementsByTagName("body")[0];
			undoStep[UNDO_NAME]=name;
			undoStep[UNDO_NODE]=getNode.cloneNode(true);
		}
		else {
				// right now this clones the document element, but when that element is replaced Mozilla's scrollbars get broken. So there has to be a better way...
			mozileDebug(f,3,"XML context");
			getNode=mozileTarget.document.documentElement;
			undoStep[UNDO_NAME]=name;
			undoStep[UNDO_NODE]=getNode.cloneNode(true);
		}
		//undoStep['position']=getPosition(getNode);
		mozileTarget.undoStack=mozileTarget.undoStack.slice(0,mozileTarget.undoCounter);			// make sure there are no extra steps
		if(mozileTarget.undoStack.length > undoMax) {
			mozileTarget.undoStack=mozileTarget.undoStack.slice(1);			// make sure there are no extra steps
		}
		mozileTarget.undoStack.push(undoStep);
		mozileTarget.undoCounter=mozileTarget.undoStack.length;
		//alert("Undo:"+mozileTarget.undoStack.length +":"+mozileTarget.undoCounter);
		mozileUpdateUndoButtons();
	} catch(e) {
		mozileDebug(f,1,"Error with the document:" + e);
	}
} // end mozileDo


// Update Undo Buttons
// This works but is not currently used.
// It gives the user some feedback about when more undo/redo steps are available.
//
function updateUndoButtons() {
	var f = ["mozileUndoRedo.js","updateUndoButtons"];
	var u=document.getElementById('mozile-undo');
	var r=document.getElementById('mozile-redo');
	try {
  	if(mozileTarget.undoCounter<1) {
  		u.setAttribute('disabled','true');
  	}else {
  		u.removeAttribute('disabled');
  	}
  	if(mozileTarget.undoCounter>mozileTarget.undoStack.length-2) {
  		r.setAttribute('disabled','true');
  	}else {
  		r.removeAttribute('disabled');
  	}
	}
	catch (e) {
		mozileDebug(f,1,"Exception! "+e);
	}
} // end updateUndoButtons()

// Dump Undo Stack
// Mostly for debugging

function dumpUndoStack() {
	var out="Current: "+mozileTarget.undoCounter+"\n";
	for(var i=0;i<mozileTarget.undoStack.length;i++) {
		cur=mozileTarget.undoStack[i];
		out+="["+i+"] "+cur['name'] +": "+ cur['node'].tagName +"\n";
	}
	alert(out);
} // end dumpUndoStack()




// Mozile Undo
// Moves doen the undoStack from the curent undoCounter,
// replacing the current BODY node with the one from the stack.
// Behaves a bit strangely around the undoMax limit.

function mozileUndo() {
	if( (("undoStack" in mozileTarget) != true) || (!mozileTarget.undoStack) ) {
		mozileTarget.undoStack=new Array();
		mozileTarget.undoCounter=0;
		return;
	}
	var undoStack=mozileTarget.undoStack;				// it is safe to copy this, since it will not be manipulated

	if(mozileTarget.undoCounter<1) {
		//alert("Nothing to Undo!");
		return;
	}
	if(mozileTarget.undoCounter==undoStack.length) {
		//alert("Current");
		mozileDo('Current');
 		mozileTarget.undoCounter--;
	}
 	mozileTarget.undoCounter--;
	var undoStep=undoStack[mozileTarget.undoCounter];

	mozileReplaceNode(undoStep[UNDO_NODE]);
	mozileUpdateUndoButtons();
} // end mozileUndo()



// Mozile Redo
// Moves up the undoStack from the current undoCounter,
// replacing the current BODY noe with the one in the stack.

function mozileRedo() {
	if(!mozileTarget.undoStack) {
		mozileTarget.undoStack=new Array();
		mozileTarget.undoCounter=0;
		return;
	}
	var undoStack=mozileTarget.undoStack;				// it is safe to copy this, since it will not be manipulated

	if(mozileTarget.undoCounter>undoStack.length-2) {
		//alert("Nothing to Redo!");
		return;
	}
 	mozileTarget.undoCounter++;
	var undoStep=undoStack[mozileTarget.undoCounter];

	mozileReplaceNode(undoStep[UNDO_NODE]);
	mozileUpdateUndoButtons();
} // end mozileRedo()


	// Replace a special target node with the replacement.
	// The target will depend on the content, for HTML it is the body tag,
	// for XML it is currently the document element, despite problems...
function mozileReplaceNode(replacement) {
	var f = ["mozileUndoRedo.js","mozileReplaceNode"];	
	mozileDebug(f,3,"Replacing Node "+ replacement);
	var target;
	if(mozileTarget.document.documentElement.tagName=="HTML") {
		mozileDebug(f,3,"HTML context");
		target = mozileTarget.document.getElementsByTagName("body")[0];
		try {
			mozileTarget.document.documentElement.replaceChild(replacement,target);
		}catch (e) {
			mozileDebug(f,1,"Exception! "+e);
		}
	}
	else {
		mozileDebug(f,3,"XML context");
		target = mozileTarget.document.documentElement;
		try {
			mozileTarget.document.replaceChild(replacement,target);
		}catch (e) {
			mozileDebug(f,1,"Exception! "+e);
		}
	}
}

